"""DAG definition and node implementations."""
import os
import json
import logging
from string import Template
import yaml
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt
from rich import print
from typing import Dict, Any, Type, Callable
from .nodes import Node, NodeRegistry
from .backends import Backend, BackendRegistry
from .default_functions import read_sql, llm_filter, llm_call

log: logging.Logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def create_node(
    node_config: Dict[str, Any],
    secrets: Dict[str, str],
    runtime_config_params: Dict[str, str] = None,
    custom_functions: Dict[str, Callable] = None,
) -> Node:
    node_type = node_config.get("type")
    node_class = NodeRegistry.get(node_type)
    if node_class is None:
        raise ValueError(
            f"Unknown node type: `{node_type}`. You likely forgot to define its `type`, have it listed as an input somewhere without first creating the node, or haven't given a custom node."
        )

    node = node_class(
        node_config,
        secrets,
        runtime_config_params=runtime_config_params,
        custom_functions=custom_functions
    )
    node.validate_config()
    return node


def _is_json_serializable(data: Any) -> bool:
    """Check if the data is JSON serializable, or is a DataFrame which
    is JSON serializable."""
    try:
        if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
            data.to_json(orient="records")
        elif hasattr(data, "to_json"):
            data.to_json()
        elif isinstance(data, dict):
            # Check keys for DataFrames.
            for key, value in data.items():
                if isinstance(value, pd.DataFrame):
                    value.to_json(orient="records")
                elif not _is_json_serializable(value):
                    return False
        elif isinstance(data, list):
            is_all_dataframes_or_dicts = all(
                isinstance(item, (pd.DataFrame, dict)) for item in data
            )
            if is_all_dataframes_or_dicts:
                for item in data:
                    if isinstance(item, pd.DataFrame):
                        item.to_json(orient="records")
                    else:
                        json.dumps(item)
        else:
            json.dumps(data)
        return True
    except TypeError:
        return False


"""DAG definition."""


class WordcelDAG:
    """DAG class to define and execute a directed acyclic graph."""

    default_functions = {
        "read_sql": read_sql,
        "llm_call": llm_call,
        "llm_filter": llm_filter,
    }

    def __init__(
        self,
        yaml_file: str,
        secrets_file: str = None,
        runtime_config_params: Dict[str, str] = None,
        custom_functions: Dict[str, Callable] = None,
        custom_nodes: Dict[str, Type[Node]] = None,
        custom_backends: Dict[str, Type[Backend]] = None,
    ):
        """
        Initialize the DAG from a YAML file.
        @param yaml_file: The path to the YAML file containing the DAG
        configuration.
        @param secrets_file: The path to the YAML file containing the
        secrets.
        @param runtime_config_params: A dictionary of parameters to substitute
        in the YAML file at runtime.
        @param custom_functions: A dictionary of custom functions to register.
        @param custom_nodes: A dictionary of custom nodes to register.
        @param custom_backends: A dictionary of custom backends to register.
        """
        log.warning("This class is still experimental: use at your own risk.")

        # First load the configuration and secrets.
        self.runtime_config_params = runtime_config_params
        if self.runtime_config_params:
            # If we are provided with config params, substitute them in the YAML file.
            with open(yaml_file, "r") as f:
                pipeline_content = f.read()
            yaml_file = Template(pipeline_content).safe_substitute(
                self.runtime_config_params
            )
        self.config = WordcelDAG.load_yaml(yaml_file)
        self.name = self.config["dag"]["name"]
        self.backend_config = self.config["dag"].get("backend", {})
        self.secrets = {}
        if secrets_file is not None:
            self.secrets = WordcelDAG.load_secrets(secrets_file)

        # Register functions, nodes, backend.
        self._register_default_and_custom_functions(custom_functions)
        self._register_default_and_custom_nodes(custom_nodes)
        self._register_default_and_custom_backends(custom_backends)

        # Create the backend.
        backend_type = self.backend_config.get("type") if self.backend_config else None
        self.backend = None
        if backend_type:
            assert (
                BackendRegistry.get(backend_type) is not None
            ), f"Unknown backend: {backend_type}"
            self.backend = BackendRegistry.get(backend_type)(self.backend_config)

        # Create the graph and nodes.
        self.graph = self.create_graph()
        self.nodes = self.create_nodes()

    @staticmethod
    def load_yaml(yaml_file: str) -> Dict[str, Any]:
        """Load a YAML (or JSON) file."""
        yaml_file = os.path.expanduser(yaml_file)
        if yaml_file.endswith(".json"):
            with open(yaml_file, "r") as file:
                return json.load(file)
        elif yaml_file.endswith(".yaml"):
            with open(yaml_file, "r") as file:
                return yaml.safe_load(file)
        elif isinstance(yaml_file, str):
            # Attempt to read the string as a yaml file.
            result = yaml.safe_load(yaml_file)
            if not isinstance(result, dict):
                result = json.loads(yaml_file)
                if not isinstance(result, dict):
                    raise ValueError(f"Unknown file type: {yaml_file}")
            return result
        else:
            raise ValueError(f"Unknown file type: {yaml_file}")
    

    @staticmethod
    def load_secrets(secrets_file: str) -> Dict[str, str]:
        """Load a secrets file."""
        return WordcelDAG.load_yaml(secrets_file)

    def _register_default_and_custom_functions(
        self, custom_functions: Dict[str, Callable] = None
    ) -> None:
        """Register custom functions."""
        self.functions = self.default_functions
        if custom_functions:
            # Assert that the custom functions do not override the defaults.
            for key, value in custom_functions.items():
                if key in self.default_functions:
                    raise ValueError(
                        f"Custom function {key} overrides default function."
                    )

            self.functions.update(custom_functions)

    def _register_default_and_custom_nodes(
        self, custom_nodes: Dict[str, Type[Node]] = None
    ) -> None:
        """Register custom nodes."""
        NodeRegistry.register_default_nodes()
        if custom_nodes:
            for node_type, node_class in custom_nodes.items():
                NodeRegistry.register(node_type, node_class)

    def _register_default_and_custom_backends(
        self, custom_backends: Dict[str, Type[Backend]] = None
    ) -> None:
        """Register custom backends."""
        BackendRegistry.register_default_backends()
        if custom_backends:
            for backend_type, backend_class in custom_backends.items():
                BackendRegistry.register(backend_type, backend_class)

    def save_image(self, path: str) -> None:
        """Save an image of the DAG using graph.draw."""
        subset_key = "__wordcel_dag_layer__"
        # `multipartite_layout` expects the layer as a node attribute, so add the
        # numeric layer value as a node attribute
        for layer, node in enumerate(nx.topological_generations(self.graph)):
            for n in node:
                self.graph.nodes[n][subset_key] = layer

        pos = nx.multipartite_layout(self.graph, subset_key=subset_key)
        nx.draw_networkx(self.graph, with_labels=True, pos=pos)
        # Make the image LARGE!
        plt.gcf().set_size_inches(18.5, 10.5)
        plt.savefig(path)
        plt.close()

    def create_graph(self) -> nx.DiGraph:
        G = nx.DiGraph()
        for node in self.config["nodes"]:
            G.add_node(node["id"], **node)
            if "input" in node:
                if isinstance(node["input"], list):
                    for input_node in node["input"]:
                        G.add_edge(input_node, node["id"])
                else:
                    G.add_edge(node["input"], node["id"])
        return G

    def create_nodes(self) -> Dict[str, Node]:
        """Create node instances from the graph configuration."""
        nodes = {}
        for node_id, node_config in self.graph.nodes(data=True):
            try:
                nodes[node_id] = create_node(
                    node_config,
                    self.secrets,
                    runtime_config_params=self.runtime_config_params,
                    custom_functions=self.default_functions
                )
            except ValueError as e:
                raise ValueError(f"Error creating node {node_id}: {str(e)}")
        return nodes


    def __prepare_incoming_input(self, input_data, results, node_id: str):
        """Prepare the incoming input for a node."""
        incoming_edges = self.graph.nodes[node_id].get("input")
        incoming_input = None
        if input_data and isinstance(input_data, dict) and node_id in input_data:
            # First check if the input data is given at runtime.
            # If so, we don't need to look at the incoming edges.
            assert (
                incoming_edges is None
            ), "Node cannot have both `input` and input data given at runtime."
            incoming_input = input_data[node_id]
        elif isinstance(incoming_edges, list):
            incoming_input = [results[input_id] for input_id in incoming_edges]
        else:
            incoming_input = results.get(incoming_edges)

        return incoming_input
    
    def __check_result_is_json_serializable(self, results, node_id: str):
        if not _is_json_serializable(results[node_id]):
            result_type = type(results[node_id]).__name__
            raise ValueError(
                f"Node `{node_id}` returned type `{result_type}`, which either is not serializable or contains something, like a DataFrame, that is not serializable."
            )

    def execute(self, input_data: Dict[str, Any] = None, verbose=False) -> Dict[str, Any]:
        """Execute the DAG.

        @param input_data: A dictionary of input data for the nodes. The key
        is the node ID that the input data is for.
        """
        # Sort and execute the nodes.
        results = {}

        for node_id in nx.topological_sort(self.graph):
            log.info(f"Executing node `{node_id}` from DAG `{self.name}`.")
            node = self.nodes[node_id]

            # Get the incoming edges and their inputs.
            incoming_input = self.__prepare_incoming_input(input_data, results, node_id)
            
            # Check the cache, if we have a backend.
            if self.backend and self.backend.exists(node_id, incoming_input):
                log.info(f"Loading node `{node_id}` from cache.")
                results[node_id] = self.backend.load(node_id, incoming_input)
            else:
                results[node_id] = node.execute(incoming_input)

                # If the node is not a DAG node, check if the result is JSON serializable.
                is_dag_node = isinstance(node, NodeRegistry.get("dag"))
                if not is_dag_node:
                    self.__check_result_is_json_serializable(results, node_id)

                # Don't save DAG nodes to cache, since they may have their own cache.
                if self.backend and not is_dag_node:
                    log.info(f"Saving node `{node_id}` to cache.")
                    self.backend.save(node_id, incoming_input, results[node_id])

            if verbose:
                print(f"Result for node {node_id}:")
                print(results[node_id])

        return results



"""CLI for Wordcel."""

import os
import logging
import click
from rich import print
from wordcel.dag.utils import initialize_dag

log: logging.Logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

PIPELINE_TEMPLATE = """
dag:
  name: "Your DAG Name"
#   backend:
#     type: local
#     cache_dir: path/to/your/cache/folder

nodes:
  - id: get_data
    type: csv
    path: "https://raw.githubusercontent.com/cs109/2014_data/master/countries.csv"

  - id: df_filter
    type: dataframe_operation
    input: get_data
    operation: "head"
    args: [2]

  - id: llm_filter
    input: df_filter
    type: llm_filter
    column: "Country"
    prompt: "Is this country in Africa? Answer only Yes or No."

  - id: process_filtered
    type: llm
    template: "What cuisine is this country known for? {input}"
    input: llm_filter
    input_field: "Country"
    output_field: "Cuisine"

#   - id: save_results
#     type: file_writer
#     path: "example_output.txt"
#     input: process_filtered

# Other node types you might like:

# - id: load_json_data
#   type: json
#   path: /path/to/your/data.json

# - id: load_json_data
#   type: json
#   path: /path/to/your/data.json

# - id: load_json_as_dataframe
#   type: json_dataframe
#   path: /path/to/your/data.json
#   read_json_kwargs:
#     orient: records

# - id: format_string
#   type: string_template
#   template: "Hello, ${name}! You are ${age} years old."
#   input: previous_node_id

# - id: run_custom_script
#   type: python_script
#   script_path: /path/to/your/script.py
#   args: ["arg1", "arg2"]

# - id: sub_dag
#   type: dag
#   path: /path/to/sub_dag.yaml
#   secrets_path: /path/to/sub_dag_secrets.yaml

"""


@click.group()
def main():
    """Wordcel CLI."""
    pass


@main.group()
def dag():
    """WordcelDAG commands."""
    pass


@dag.command()
@click.argument("pipeline_file")
def new(pipeline_file):
    """Create a new pipeline."""
    if os.path.exists(pipeline_file):
        click.echo(f"Pipeline file {pipeline_file} already exists.")
        return

    with open(pipeline_file, "w") as f:
        f.write(PIPELINE_TEMPLATE)

    click.echo(f"Pipeline file {pipeline_file} created.")


@dag.command()
def list_node_types():
    """List available node types."""
    from wordcel.dag.nodes import NODE_TYPES

    for node_type, node_class in NODE_TYPES.items():
        click.echo(f"- {node_type}: {node_class.description}")


@dag.command()
@click.argument("pipeline_file")
@click.argument("save_path")
@click.option("--custom-nodes", default=None, help="Path to custom nodes Python file.")
@click.option("--secrets", default=None, help="Path to secrets file.")
def visualize(pipeline_file, save_path, custom_nodes, secrets):
    """Visualize a pipeline."""
    dag = initialize_dag(pipeline_file, secrets_file=secrets, custom_nodes=custom_nodes)
    log.info("Saving visualization to %s.", save_path)
    dag.save_image(save_path)


@dag.command()
@click.argument("pipeline_file")
@click.option("--secrets", default=None, help="Path to secrets file.")
@click.option(
    "--custom-nodes",
    default=None,
    multiple=True,
    help="Path to custom nodes Python file.",
)
@click.option(
    "--custom-functions",
    default=None,
    multiple=True,
    help="Path to custom functions Python file.",
)
@click.option(
    "--custom-backends",
    default=None,
    multiple=True,
    help="Path to custom backends Python file.",
)
@click.option("--visualization", default=None, help="Path to save visualization.")
@click.option("--verbose", "-v", is_flag=True, help="Verbose output.")
@click.option(
    "--input",
    "-i",
    nargs=2,
    type=(str, str),
    multiple=True,
    help="Input data for the pipeline. Given in the format `-i node_id value`.",
)
@click.option(
    "--config-param",
    "-c",
    nargs=2,
    type=(str, str),
    multiple=True,
    help="Substitute values into the config file at runtime. Given in the format `-c param value`.",
)
def execute(
    pipeline_file,
    secrets,
    custom_nodes,
    custom_functions,
    custom_backends,
    visualization,
    verbose,
    input,
    config_param,
):
    """Execute a pipeline."""
    from rich.console import Console
    from rich.panel import Panel
    from rich.tree import Tree

    
    dag = initialize_dag(
        pipeline_file,
        config_params=dict(config_param),
        secrets_file=secrets,
        custom_nodes=custom_nodes,
        custom_functions=custom_functions,
        custom_backends=custom_backends
    )

    if input:
        print("Given input data: ", dict(input))

    results = dag.execute(input_data=dict(input), verbose=verbose)

    console = Console()
    tree = Tree("⚡️ [bold blue]Execution Results")

    for node_id, result in results.items():
        node_tree = tree.add(f"[bold green]{node_id}")
        node_tree.add(f"{result}")

    console.print(Panel(tree, expand=False, border_style="bold"))

    if visualization:
        dag.save_image(visualization)


if __name__ == "__main__":
    main()



"""Default helper functions for the DAG / Nodes."""
import pandas as pd
from sqlalchemy import create_engine
from ..featurize import apply_io_bound_function
from ..llms import llm_call


def read_sql(query: str, connection_string: str) -> pd.DataFrame:
    """Helper function to execute a read-only SQL query."""
    engine = create_engine(connection_string)
    results = pd.read_sql(query, connection_string)
    engine.dispose()
    return results


def llm_filter(
    df: pd.DataFrame, column: str, prompt: str, model="gpt-4o-mini", num_threads: int = 1
) -> pd.DataFrame:
    """Helper function to filter a DataFrame using an LLM yes/no question."""
    if num_threads == 1:
        results = df[column].apply(
            lambda value: llm_call(
                prompt + "\n\n----\n\n" + value, model=model
            )
        )
    else:
        results = apply_io_bound_function(
            df,
            lambda value: llm_call(
                prompt + "\n\n----\n\n" + value, model=model
            ),
            text_column=column,
            num_threads=num_threads,
        )

    results = results.str.lower().str.strip()
    return df[results.str.lower().str.startswith("yes")]


"""Node definitions."""
import os
import ast
import glob
import json
import shlex
import yaml
import subprocess
from string import Template
import pandas as pd
import logging
import concurrent.futures
from abc import ABC, abstractmethod
from typing import Dict, Any, Type, Callable, List, Union
from .default_functions import read_sql, llm_filter

log: logging.Logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class Node(ABC):
    def __init__(
        self,
        config: Dict[str, Any],
        secrets: Dict[str, str],
        runtime_config_params = None,
        custom_functions: Dict[str, Callable] = None,
    ):
        self.config = config
        self.secrets = secrets
        self.functions = {}
        self.runtime_config_params = runtime_config_params
        if custom_functions:
            self.functions.update(custom_functions)


    @abstractmethod
    def execute(self, input_data: Any) -> Any:
        """
        Execute the node's operation.

        :param input_data: The input data for this node, typically the output from the previous node.
        :return: The result of this node's operation.
        """
        pass

    @abstractmethod
    def validate_config(self) -> bool:
        """
        Validate the node's configuration.

        :return: True if the configuration is valid, False otherwise.
        """
        pass


class CSVNode(Node):
    description = """Node to read a CSV file."""

    def execute(self, input_data: Any) -> pd.DataFrame:
        path = os.path.expanduser(self.config["path"])
        return pd.read_csv(path)

    def validate_config(self) -> bool:
        assert "path" in self.config, "CSVNode must have a 'path' configuration."
        return True
    

class YAMLNode(Node):
    description = """Node to read a YAML file."""

    def execute(self, input_data: Any) -> Dict[str, Any]:
        path = os.path.expanduser(self.config["path"])
        with open(path, "r") as file:
            return yaml.safe_load(file)

    def validate_config(self) -> bool:
        assert "path" in self.config, "YAMLNode must have a 'path' configuration."
        return True


class JSONNode(Node):
    description = """Node to read a JSON file."""

    def execute(self, input_data: Any) -> Dict[str, Any]:
        path = os.path.expanduser(self.config["path"])
        with open(path, "r") as file:
            return json.load(file)

    def validate_config(self) -> bool:
        assert "path" in self.config, "JSONNode must have a 'path' configuration."
        return True


class JSONDataFrameNode(Node):
    description = """Node to read a JSON file into a pandas DataFrame."""

    def execute(self, input_data: Any) -> pd.DataFrame:
        path = os.path.expanduser(self.config["path"])
        return pd.read_json(path, **self.config.get("read_json_kwargs", {}))

    def validate_config(self) -> bool:
        assert "path" in self.config, "JSONDataFrameNode must have a 'path' configuration."
        return True


class FileDirectoryNode(Node):
    description = """Node to read text and markdown files from a directory or list of directories, supporting regex patterns."""

    def execute(self, input_data: Any) -> pd.DataFrame:
        paths = self.config.get("path")
        if isinstance(input_data, dict) and "path" in input_data:
            paths = input_data["path"]
            assert "path" not in self.config, "FileDirectoryNode `path` cannot be in input data if it is in the configuration."
            if isinstance(paths, list):
                assert all(isinstance(path, str) for path in paths), "FileDirectoryNode `path` in input data must be a list of strings."
        if isinstance(paths, str):
            paths = [paths]
        
        paths = [os.path.expanduser(path) for path in paths]

        file_contents = []
        for path in paths:
            for file_path in glob.glob(path, recursive=True):
                if file_path.lower().endswith(('.txt', '.md', '.html')):
                    with open(os.path.expanduser(file_path), 'r', encoding='utf-8') as file:
                        content = file.read()
                        file_contents.append({
                            'file_path': file_path,
                            'content': content,
                            'file_type': os.path.splitext(file_path)[1][1:]  # Get file extension without the dot
                        })
        
        return pd.DataFrame(file_contents)

    def validate_config(self) -> bool:
        if "path" in self.config:
            paths = self.config["path"]
            assert isinstance(paths, str) or isinstance(paths, list), "FileDirectoryNode 'path' must be a string or a list of strings."
        return True


class SQLNode(Node):
    description = """Node to execute a SQL query."""

    def execute(self, input_data: Any) -> pd.DataFrame:
        connection_string = self.secrets["database_url"]
        read_sql_fn = self.functions.get("read_sql", read_sql)
        return read_sql_fn(self.config["query"], connection_string)

    def validate_config(self) -> bool:
        assert (
            "query" in self.config and "database_url" in self.secrets
        ), "SQLNode must have a 'query' configuration and a 'database_url' secret."
        return True


class StringTemplateNode(Node):
    description = """Node to apply a string template to input data. The
    `header`, if given, is placed at the top. The `template` is filled in
    at least once, and potentially many times depending on the input."""

    def execute(self, input_data: Union[Dict[str, Any], List[Dict[str, Any]], pd.DataFrame]) -> Union[str, List[str]]:
        template = Template(self.config["template"])
        mode = self.config.get("mode", "single")

        return_string = self.config.get("header", "")
        if return_string:
            return_string = return_string + "\n\n"

        if isinstance(input_data, (list, pd.DataFrame)):
            records = input_data
            if isinstance(input_data, pd.DataFrame):
                records = input_data.to_dict(orient="records")

            if mode == "single":  # Single string output.
                for item in records:
                    if isinstance(item, dict):
                        return_string = return_string + template.safe_substitute(item) + "\n\n"
                    elif isinstance(item, str):
                        return_string = return_string + template.safe_substitute(input=item) + "\n\n"
            elif mode == "multiple":  # List of strings output.
                return_items = [
                    template.safe_substitute(item)
                    if isinstance(item, dict)
                    else template.safe_substitute(input=item)
                    for item in records
                ]
                return return_items
        elif isinstance(input_data, dict):
            return_string = return_string + template.safe_substitute(input_data)
        elif isinstance(input_data, str):
            return_string = return_string + template.safe_substitute(input=input_data)
        elif input_data is None:
            return_string = return_string + template.safe_substitute()
    
        return return_string

    def validate_config(self) -> bool:
        assert "template" in self.config, "StringTemplateNode must have a 'template' configuration."
        if "mode" in self.config:
            modes_are_valid = self.config["mode"] in ["single", "multiple"]
            assert modes_are_valid, "StringTemplateNode `mode` must be `single` or `multiple`."
        return True


class LLMNode(Node):
    description = """Node to query an LLM API with a template. Template must
    contain "{input}" in order to substitute something in. If given a string,
    it will fill in the template and return the result. If given a DataFrame
    or list of dicts, it will turn the column / field denoted by `input_field`
    into a list of strings, fill in the template for each string."""

    def _try_to_load_as_json(self, text: str) -> Union[str, dict]:
        try:
            return json.loads(text)
        except json.JSONDecodeError:
            return text

    def execute(
        self, input_data: Union[str, List[str], pd.DataFrame]
    ) -> Union[str, List[str]]:
        llm_call = self.functions["llm_call"]
        num_threads = self.config.get("num_threads", 1)
        assert num_threads >= 1, "Number of threads must be at least 1."
        model = self.config.get("model", "gpt-4o-mini")
        if num_threads > 1:
            log.info(f"Using {num_threads} threads for LLM Node.")

        # If it's a single string, just call the LLM once.
        if isinstance(input_data, str):
            response = llm_call(
                self.config["template"].format(input=input_data),
                model=model
            )
            return self._try_to_load_as_json(response)
        # If it is a single dict, call the LLM once.
        elif isinstance(input_data, dict):
            assert "input_field" in self.config, "LLMNode must have a `input_field` configuration if given a dict."
            response = llm_call(
                self.config["template"].format(input=input_data[self.config["input_field"]]),
                model=model
            )
            return self._try_to_load_as_json(response)
        
        # Turn input_data into a list of strings.
        if isinstance(input_data, pd.DataFrame):
            # Assert that `input_field` is in the configuration.
            assert "input_field" in self.config, "LLMNode must have a `input_field` configuration for DataFrame or list of dicts."
            # Turn it into a list of strings.
            texts = input_data[self.config["input_field"]].tolist()
        elif isinstance(input_data, pd.Series):
            texts = input_data.tolist()
        elif isinstance(input_data, list):
            is_all_strings = all(isinstance(item, str) for item in input_data)
            is_all_dicts = all(isinstance(item, dict) for item in input_data)
            is_all_dataframes = all(isinstance(item, pd.DataFrame) for item in input_data)

            assert is_all_strings or is_all_dicts or is_all_dataframes, "LLMNode input must be a list of strings, dicts, or DataFrames."
            if is_all_strings:
                texts = input_data
            elif is_all_dicts:
                texts = [item[self.config["input_field"]] for item in input_data]
            else:
                assert "input_field" in self.config, "LLMNode must have a `input_field` configuration for list of DataFrames."
                texts = []
                for df in input_data:
                    texts.extend(df[self.config["input_field"]].tolist())

        # Call the LLM in parallel.
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=num_threads
        ) as executor:
            results = list(
                executor.map(
                    lambda text: llm_call(
                        self.config["template"].format(input=text),
                        model=model
                    ),
                    texts,
                )
            )
        results = [self._try_to_load_as_json(result) for result in results]

        # Now reshape the output to be compatible with the input.
        if isinstance(input_data, pd.DataFrame):
            new_df = input_data.copy()
            new_df.loc[:, self.config["output_field"]] = results
            return new_df
        elif isinstance(input_data, pd.Series):
            return results  # Just keep it as a list.
        elif isinstance(input_data, list):
            if is_all_strings:  # Case: list of strings.
                return results
            elif is_all_dicts:  # Case: list of dicts.
                return [
                    {**item, self.config["output_field"]: result}
                    for item, result in zip(input_data, results)
                ]
            else:  # Case: list of DataFrames.
                dfs = []
                for df in input_data:
                    new_df = df.copy()
                    new_df.loc[:, self.config["output_field"]] = results[: len(df)]
                    results = results[len(df) :]
                    dfs.append(df)
                return dfs
        
    def validate_config(self) -> bool:
        assert (
            "template" in self.config
        ), "LLMNode must have a 'template' configuration."
        assert (
            "input" in self.config
        ), "LLMNode must have an 'input' configuration."
        assert "{input}" in self.config["template"], "LLMNode template must contain '{input}'."

        # If `input_field` in the configuration, then there must also be an `output_field`.
        if "input_field" in self.config:
            assert "output_field" in self.config, "LLMNode must have an `output_field` configuration if `field` is present."
        return True


class LLMFilterNode(Node):
    description = """Node to use LLMs to filter a dataframe."""

    def execute(self, input_data: pd.DataFrame) -> Union[pd.DataFrame, pd.Series]:
        is_dataframe = isinstance(input_data, pd.DataFrame)
        assert is_dataframe, "LLMFilterNode must have a DataFrame as input."
        model = self.config.get("model", "gpt-4o-mini")
        num_threads = self.config.get("num_threads", 1)
        assert num_threads >= 1, "Number of threads must be at least 1."
        if num_threads > 1:
            log.info(f"Using {num_threads} threads for LLMFilter Node.")

        llm_filter_fn = self.functions.get("llm_filter", llm_filter)
        return llm_filter_fn(
            input_data,
            self.config["column"],
            self.config["prompt"],
            model=model,
            num_threads=num_threads,
        )

    def validate_config(self) -> bool:
        assert (
            "column" in self.config and "prompt" in self.config
        ), "LLMFilterNode must have 'column' and 'prompt' configuration."
        assert (
            "input" in self.config
        ), "LLMFilterNode must have an 'input' configuration."
        assert isinstance(
            self.config["input"], str
        ), "LLMFilter node 'input' configuration must have only one input."
        return True


class FileWriterNode(Node):
    description = """Node to write data to a file."""

    def execute(self, input_data: str) -> str:
        mode = self.config.get("mode", "single")
        if mode == "single":
            path = os.path.expanduser(self.config["path"])
            with open(path, "w") as file:
                file.write(str(input_data))
        elif mode == "multiple":
            for i, data in enumerate(input_data):
                path = os.path.expanduser(self.config["path"].format(i=i))
                with open(path, "w") as file:
                    file.write(str(data))

    def validate_config(self) -> bool:
        assert (
            "path" in self.config
        ), "FileWriterNode must have a 'path' configuration."
        if "mode" in self.config:
            modes_are_allowed = self.config["mode"] in ["single", "multiple"]
            assert modes_are_allowed, "FileWriterNode `mode` must be `single` or `multiple`."

            if self.config["mode"] == "multiple":
                assert "{i}" in self.config["path"], "FileWriterNode `path` must contain \"{i}\" to format for `multiple` mode."
        return True


class DataFrameOperationNode(Node):
    description = """Node to apply a DataFrame operation to the input data."""

    def execute(
        self, input_data: Union[pd.DataFrame, List]
    ) -> pd.DataFrame:
        if not isinstance(input_data, list):
            input_data = [input_data]
        return self._apply_operation(input_data)
        
    def __handle_dataframe_method(self, df: pd.DataFrame, operation: str, *args, **kwargs) -> pd.DataFrame:
        """Handle DataFrame methods."""
        if hasattr(df, operation):
            method = getattr(df, operation)
            if callable(method):
                try:
                    return method(*args, **kwargs)
                except TypeError:
                    raise TypeError(f"Error calling method {operation} with args {args} and kwargs {kwargs}.")
            else:
                return method
        else:
            raise ValueError(f"Unknown DataFrame operation: {operation}")

    def _apply_operation(self, input_array: List) -> pd.DataFrame:
        operation = self.config["operation"]
        args = self.config.get("args", [])
        kwargs = self.config.get("kwargs", {})

        every_element_is_dataframe = all(isinstance(df, pd.DataFrame) for df in input_array)

        if operation == "concat":
            assert every_element_is_dataframe, "All inputs must be DataFrames."
            return pd.concat(input_array, **kwargs)
        elif operation == "merge":
            assert len(input_array) == 2, "Merge operation requires exactly two DataFrames."
            assert every_element_is_dataframe, "All inputs must be DataFrames."
            return pd.merge(input_array[0], input_array[1], **kwargs)
        elif operation == "set_column":
            # Assert that one input is a DataFrame and the other is a string, list, or Series.
            assert len(input_array) == 2, "`set_column` operation requires exactly two inputs."
            assert isinstance(input_array[0], pd.DataFrame), "First input must be a DataFrame."
            assert isinstance(input_array[1], (str, list, pd.Series)), "Second input must be a string, list, or Series."
            df = input_array[0]
            df.loc[:, self.config["column_name"]] = input_array[1]
            return df
        else:
            # Apply the operation to each DataFrame.
            completed = [
                self.__handle_dataframe_method(df, operation, *args, **kwargs)
                for df in input_array
            ]
            if len(completed) == 1:
                return completed[0]
            return completed

    def validate_config(self) -> bool:
        assert (
            "operation" in self.config
        ), "DataFrameOperationNode must have an `operation` configuration."
        if self.config["operation"] == "set_column":
            assert (
                "column_name" in self.config
            ), "DataFrameOperationNode with `set_column` operation must have a `column_name` configuration."
        return True


class PythonScriptNode(Node):
    description = """Node to execute a Python script using subprocess."""

    def execute(self, input_data: Any) -> Any:
        if input_data is not None:
            if isinstance(input_data, pd.Series):
                input_data = input_data.tolist()
            elif isinstance(input_data, (str, int, float, complex, bool)):
                input_data = [input_data]
            elif isinstance(input_data, list):
                input_data = input_data
            else:
                raise ValueError("Input data must be a primitive, list, or pandas Series.")
        else:
            # None if no input data, so that it runs once and appends
            # nothing to the command.
            input_data = [None]

        # Prepare the command.
        script_path = shlex.quote(self.config["script_path"])
        args = self.config.get("args", [])
        command = ["python", script_path]
        if args:
            # Add args from YAML config.
            for arg in args:
                if isinstance(arg, dict):
                    for k, v in arg.items():
                        command.extend([f"--{shlex.quote(k)}", shlex.quote(str(v))])
                else:
                    command.append(shlex.quote(str(arg)))

        # Prepare the results list and execute the script.
        results = []
        for item in input_data:
            cmd = command.copy()
            if item is not None:
                cmd.append(shlex.quote(str(item)))

            # Execute the script.
            log.info("Attempting to execute command: " + shlex.join(cmd))
            try:
                subprocess_result = subprocess.run(cmd, shell=False, stdout=subprocess.PIPE)
            except subprocess.CalledProcessError as e:
                raise RuntimeError(f"Script execution failed: {e.stderr}")
        
            # Handle output.
            output_file_key = "return_output_file"
            return_stdout_key = "return_stdout"
            if output_file_key in self.config:
                assert self.config[output_file_key].endswith(
                    (".csv", ".json")
                ), f"{output_file_key} must be a csv or json file."

                # Read the file and return the result.
                if self.config[output_file_key].endswith(".csv"):
                    result = pd.read_csv(self.config[output_file_key])
                    results.append(result)
                elif self.config[output_file_key].endswith(".json"):
                    with open(self.config[output_file_key], "r") as file:
                        result = json.load(file)
                        results.append(result)
            elif return_stdout_key in self.config and self.config[return_stdout_key]:
                result = subprocess_result.stdout.decode("utf-8").strip()
                try:
                    result = ast.literal_eval(result)
                except ValueError:
                    raise ValueError(f"Could not properly eval the `stdout` of the PythonScriptNode output for command: {cmd}. Please make sure you are printing out well-formed JSON.")
                results.append(result)

        if len(results) == 1:
            return results[0]

        return results

    def validate_config(self) -> bool:
        assert "script_path" in self.config, "PythonScript node must have a `script_path` configuration."
        assert os.path.exists(self.config["script_path"]), "PythonScript node `script_path` does not exist."
        # Assert that only one of output_file or return_stdout is present, but not both.
        has_output_file = "return_output_file" in self.config
        has_return_stdout = "return_stdout" in self.config
        not_both = has_output_file != has_return_stdout
        neither = not has_output_file and not has_return_stdout
        assert not_both or neither, "PythonScript node must have either `output_file` or `return_stdout` configuration, or neither."
        return True


class DAGNode(Node):
    description = """Node to execute a sub-DAG defined in a YAML file."""

    def execute(self, input_data: Dict[str, Any]) -> Any:
        from .dag import WordcelDAG
        # Ensure that the custom_functions don't override the default
        # functions. Subtract out the ones that do.
        self.functions = {
            key: value
            for key, value in self.functions.items()
            if key not in WordcelDAG.default_functions
        }

        # Construct the runtime_config_params. Check for conflicts between
        # the DAG instantiation, the YAML definition.
        runtime_config_params = {}
        if self.runtime_config_params:  # From DAG instantiation in code.
            runtime_config_params = self.runtime_config_params.copy()
        if "runtime_config_params" in self.config:  # From the user.
            # Check if there are any conflicts and raise an error if so.
            conflicts = set(self.config["runtime_config_params"].keys()).intersection(
                runtime_config_params.keys()
            )
            assert not conflicts, f"Runtime config params conflict between DAG instantiation and YAML definition: {conflicts}."

            # If `from_input_data` is in the runtime_config_params, then
            # we need to get the value from the input_data.
            if "from_input_data" in self.config["runtime_config_params"]:
                input_data_key = self.config["runtime_config_params"]["from_input_data"]
                # If it's not present in the input_data, then assume that the user 
                # wants to pass the entire input_data.
                if input_data_key not in input_data:
                    runtime_config_params[input_data_key] = input_data
                else:
                    runtime_config_params[input_data_key] = input_data[input_data_key]

            runtime_config_params.update(self.config["runtime_config_params"])

        # We do not need to give the custom_backends or custom_nodes to the
        # sub-DAG, as they are already in the registry.
        sub_dag = WordcelDAG(
            yaml_file=os.path.expanduser(self.config["path"]),
            secrets_file=self.config.get("secrets_path"),
            custom_functions=self.functions,
            runtime_config_params=runtime_config_params,
        )
        dag_results = sub_dag.execute(input_data=input_data)

        # The dag_results are a dict. If self.config has param `output_key`,
        # then return the value of that key.
        if self.config.get("output_key"):
            output_keys = self.config["output_key"]
            log.info(f"Returning results for output key in DAGNode: {output_keys}")
            if isinstance(output_keys, list):
                return {key: dag_results[key] for key in output_keys} 
            elif isinstance(output_keys, str):
                return dag_results[output_keys]
            else:
                raise ValueError(f"DAGNOde `output_key` must be a string or a list of strings.")
        return dag_results

    def validate_config(self) -> bool:
        assert "path" in self.config, "DAGNode must have a 'path' configuration."
        return True


NODE_TYPES: Dict[str, Type[Node]] = {
    "csv": CSVNode,
    "yaml": YAMLNode,
    "json": JSONNode,
    "json_dataframe": JSONDataFrameNode,
    "file_directory": FileDirectoryNode,
    "sql": SQLNode,
    "string_template": StringTemplateNode,
    "llm": LLMNode,
    "llm_filter": LLMFilterNode,
    "file_writer": FileWriterNode,
    "dataframe_operation": DataFrameOperationNode,
    "python_script": PythonScriptNode,
    "dag": DAGNode,
}


class NodeRegistry:
    _registry = {}

    @classmethod
    def register(cls, node_type: str, node_class: Type[Node]):
        cls._registry[node_type] = node_class

    @classmethod
    def get(cls, node_type: str) -> Type[Node]:
        return cls._registry.get(node_type)

    @classmethod
    def register_default_nodes(cls):
        for node_type, node_class in NODE_TYPES.items():
            cls.register(node_type, node_class)



"""Utility functions for the DAG module."""
import importlib.util
from typing import Dict, Type, Union, List
from rich import print
from .nodes import Node
from .backends import Backend
from .dag import WordcelDAG


def load_module(file_path, module_name):
    """Load a Python module from a file."""
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def create_custom_nodes_from_files(custom_nodes_files: Union[str, List]) -> Dict[str, Type[Node]]:
    """Create custom nodes from a Python file."""
    if isinstance(custom_nodes_files, str):
        custom_nodes_files = [custom_nodes_files]

    custom_nodes = {}
    for custom_nodes_file in custom_nodes_files:
        nodes_module = load_module(custom_nodes_file, custom_nodes_file)
        
        found_nodes = {name: cls for name, cls in nodes_module.__dict__.items() if isinstance(cls, type) and issubclass(cls, Node) and cls is not Node}
        # Check for duplicates.
        duplicates = set(custom_nodes.keys()) & set(found_nodes.keys())
        if duplicates:
            raise ValueError(f"Duplicate node(s) found in {custom_nodes_file}: {duplicates}.")

        custom_nodes.update(found_nodes)
    
    print("Created custom nodes: ", custom_nodes)
    return custom_nodes


def create_custom_functions_from_files(custom_functions_files: Union[str, List]) -> Dict[str, callable]:
    """Create custom functions from a Python file."""
    if isinstance(custom_functions_files, str):
        custom_functions_files = [custom_functions_files]

    custom_functions = {}
    for custom_functions_file in custom_functions_files:
        functions_module = load_module(custom_functions, "custom_functions")
        found_functions = {name: func for name, func in functions_module.__dict__.items() if callable(func)}
        # Check for duplicates.
        duplicates = set(custom_functions.keys()) & set(found_functions.keys())
        if duplicates:
            raise ValueError(f"Duplicate function(s) found in {custom_functions_file}: {duplicates}.")
        
        custom_functions.update(found_functions)

    print("Created custom functions: ", custom_functions)
    return custom_functions


def create_custom_backends_from_files(custom_backends_files: Union[str, List]) -> Dict[str, Type[Backend]]:
    """Create custom backends from a Python file."""
    if isinstance(custom_backends_files, str):
        custom_backends_files = [custom_backends_files]

    custom_backends = {}
    for custom_backends_file in custom_backends_files:
        backends_module = load_module(custom_backends, "custom_backends")
        found_backends = {name: cls for name, cls in backends_module.__dict__.items() if isinstance(cls, type) and issubclass(cls, Backend) and cls is not Backend}
        # Check for duplicates.
        duplicates = set(custom_backends.keys()) & set(found_backends.keys())
        if duplicates:
            raise ValueError(f"Duplicate backend(s) found in {custom_backends_file}: {duplicates}.")
        
        custom_backends.update(found_backends)

    print("Created custom backends: ", custom_backends)
    return custom_backends


def initialize_dag(
    pipeline_file,
    config_params=None,
    secrets_file=None,
    custom_nodes=None,
    custom_functions=None,
    custom_backends=None,
):
    """Initialize the DAG."""
    if custom_nodes:
        custom_nodes = create_custom_nodes_from_files(custom_nodes)

    if custom_functions:
        custom_functions = create_custom_functions_from_files(custom_functions)

    if custom_backends:
        custom_backends = create_custom_backends_from_files(custom_backends)

    dag = WordcelDAG(
        pipeline_file,
        secrets_file=secrets_file,
        runtime_config_params=config_params,
        custom_nodes=custom_nodes,
        custom_functions=custom_functions,
        custom_backends=custom_backends,
    )

    return dag


"""Backends to store data."""

import os
import hashlib
import json
import pandas as pd
from io import StringIO
from abc import ABC, abstractmethod
from typing import Any, Dict


class Backend(ABC):
    """Abstract class for a backend to store data."""

    def __init__(self, config: Dict[str, Any] = None):
        """Initialize the backend."""
        self.config = config or {}

    @abstractmethod
    def save(self, node_id: str, input_data: Any, data: Any) -> None:
        """
        Save data to the backend.
        """
        pass

    @abstractmethod
    def load(self, node_id: str, input_data: Any) -> Any:
        """
        Load data from the backend.
        """
        pass

    @abstractmethod
    def exists(self, node_id: str, input_data: Any) -> bool:
        """
        Check if data exists in the backend.
        """
        pass


class LocalBackend(Backend):
    """Local backend to store data in a cache directory."""

    def __init__(self, config: Dict[str, Any]):
        """Initialize the local backend."""
        # Check for the cache directory.
        assert "cache_dir" in config, "Local backend requires a `cache_dir`."
        cache_dir = os.path.expanduser(config.get("cache_dir"))
        if not cache_dir:
            raise ValueError("Local backend requires a `cache_dir`.")
        self.cache_dir = cache_dir
        os.makedirs(self.cache_dir, exist_ok=True)

    def _get_path(self, key: str) -> str:
        return os.path.join(self.cache_dir, f"{key}.json")

    def _generate_input_hash(self, input_data: Any) -> str:
        """Generate a hash for the input data."""
        if input_data is None:
            return "none"
        if isinstance(input_data, pd.DataFrame) or isinstance(input_data, pd.Series):
            data_str = input_data.to_json(orient="records")
        elif isinstance(input_data, (list, dict)):
            # Convert all the DataFrames or Series into JSON strings.
            if isinstance(input_data, dict):
                input_data = {
                    k: (
                        v.to_json(orient="records")
                        if isinstance(v, pd.DataFrame) or isinstance(v, pd.Series)
                        else v
                    )
                    for k, v in input_data.items()
                }
            else:
                input_data = [
                    (
                        v.to_json(orient="records")
                        if isinstance(v, pd.DataFrame) or isinstance(v, pd.Series)
                        else v
                    )
                    for v in input_data
                ]
            data_str = json.dumps(input_data, sort_keys=True)
        else:
            data_str = str(input_data)
        return hashlib.md5(data_str.encode()).hexdigest()

    def generate_cache_key(self, node_id: str, input_data: Any) -> str:
        """Generate a cache key for the node and input data."""
        input_hash = self._generate_input_hash(input_data)
        return f"{node_id}_{input_hash}"

    def save(self, node_id: str, input_data: Any, data: Any) -> None:
        """Save data or DataFrame to a JSON file."""
        if isinstance(data, pd.DataFrame) or isinstance(data, pd.Series):
            json_str = data.to_json(orient="records")
            if isinstance(data, pd.DataFrame):
                data = {"__type__": "dataframe", "data": json_str}
            elif isinstance(data, pd.Series):
                data = {"__type__": "series", "data": json_str}
        elif isinstance(data, list):
            # If it's a list of DataFrames, convert them to dicts.
            data = [
                v.to_json(orient="records") if isinstance(v, pd.DataFrame) else v
                for v in data
            ]
        elif isinstance(data, dict):
            # If it's a dict of DataFrames, convert them to dicts.
            data = {
                k: v.to_json(orient="records") if isinstance(v, pd.DataFrame) else v
                for k, v in data.items()
            }

        cache_key = self.generate_cache_key(node_id, input_data)
        with open(self._get_path(cache_key), "w") as f:
            json.dump(data, f, indent=2)

    def load(self, node_id: str, input_data: Any) -> Any:
        """Load data or DataFrame from a JSON file."""
        cache_key = self.generate_cache_key(node_id, input_data)
        with open(self._get_path(cache_key), "r") as f:
            data = json.load(f)
            # Check for the flag indicating a DataFrame.
            if data is not None:
                if isinstance(data, dict) and data["__type__"] == "dataframe":
                    data = pd.read_json(StringIO(data["data"]))
                elif isinstance(data, dict) and data["__type__"] == "series":
                    data = pd.read_json(StringIO(data["data"]), typ="series", orient="records")

            return data

    def exists(self, node_id: str, input_data: Any) -> bool:
        cache_key = self.generate_cache_key(node_id, input_data)
        return os.path.exists(self._get_path(cache_key))


BACKEND_TYPES: Dict[str, Backend] = {
    "local": LocalBackend,
}


class BackendRegistry:
    """Registry for backends."""

    _registry: Dict[str, Backend] = {}

    @classmethod
    def register(cls, name: str, backend: Backend) -> None:
        cls._registry[name] = backend

    @classmethod
    def get(cls, name: str) -> Backend:
        return cls._registry.get(name)

    @classmethod
    def register_default_backends(cls) -> None:
        for name, backend in BACKEND_TYPES.items():
            cls.register(name, backend)
